#include "jbuffer.h"
#include "sem.h"
#include <stdatomic.h>
#include <stdlib.h>

struct BNDBUF
{
  int *buffer;
  size_t size;
  volatile atomic_int start;
  volatile atomic_int end;
  SEM *empty;
  SEM *full;
};

BNDBUF *
bbCreate (size_t size)
{
  if (size == 0)
    {
      return NULL;
    }

  BNDBUF *bnd = malloc (sizeof (BNDBUF));
  if (bnd == NULL)
    {
      return NULL;
    }

  bnd->buffer = malloc (size);
  if (bnd->buffer == NULL)
    {
      free (bnd);
      return NULL;
    }

  bnd->start = 0;
  bnd->end = 0;
  bnd->size = size;

  bnd->empty = semCreate (0);
  if (bnd->empty == NULL)
    {
      free (bnd->buffer);
      free (bnd);
      return NULL;
    }

  bnd->full = semCreate (size);
  if (bnd->full == NULL)
    {
      semDestroy (bnd->empty);
      free (bnd->buffer);
      free (bnd);
    }

  return bnd;
}

void
bbDestroy (BNDBUF *bb)
{
  if (bb == NULL)
    {
      return;
    }

  semDestroy (bb->full);
  semDestroy (bb->empty);
  free (bb->buffer);
  free (bb);
}

void
bbPut (BNDBUF *bb, int value)
{
  P (bb->full);
  bb->buffer[bb->end] = value;
  bb->end = (bb->end + 1) % bb->size;
  V (bb->empty);
}

int
bbGet (BNDBUF *bb)
{
  P (bb->empty);
  int next;
  int curr;
  int ret;

  do
    {
      curr = bb->start;
      next = (curr + 1) % bb->size;
      ret = bb->buffer[curr];
    }
  while (atomic_compare_exchange_strong (&(bb->start), &curr, next) == 0);

  V (bb->full);
  return ret;
}