#include "argumentParser.h"
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

static void creeper (char *path, char *name, char type, long depth);
static void printpath (char *path, char *name);
static void die (char *msg);

int
main (int argc, char *argv[])
{
  if (initArgumentParser (argc, argv) == -1)
    {
      die ("initArgumentParser");
    }

  if (getNumberOfArguments () < 1)
    {
      fprintf (stderr, "usage: creeper path [-maxdepth=n] [-name=pattern] "
                       "[-type=d,f]\n");
      exit (EXIT_FAILURE);
    }

  char *name = getValueForOption ("name");

  char *type = getValueForOption ("type");
  if (type == NULL)
    {
      type = "0";
    }
  else if (strlen (type) != 1 || (type[0] != 'f' && type[0] != 'd'))
    {
      fprintf (stderr, "usage: creeper path [-maxdepth=n] [-name=pattern] "
                       "[-type=d,f]\n");
      exit (EXIT_FAILURE);
    }

  char *maxdepth_char = getValueForOption ("maxdepth");
  long maxdepth = -1;
  if (maxdepth_char != NULL)
    {
      errno = 0;
      char *endptr;
      maxdepth = strtol (maxdepth_char, &endptr, 10);
      if (errno != 0)
        {
          die ("strtol");
        }
      // Non empty string was fully parsed
      if (maxdepth_char == endptr || *endptr != '\0')
        {
          fprintf (stderr, "invalid maxdepth\n");
          exit (EXIT_FAILURE);
        }

      if (maxdepth < 0)
        {
          fprintf (stderr, "usage: creeper path [-maxdepth=n] [-name=pattern] "
                           "[-type=d,f]\n");
          exit (EXIT_FAILURE);
        }
    }

  for (int i = 0; i < getNumberOfArguments (); i++)
    {
      creeper (getArgument (i), name, type[0], maxdepth);
    }

  return EXIT_SUCCESS;
}

static void
creeper (char *path, char *name, char type, long depth)
{
  int fd = open (path, 0);
  if (fd == -1)
    {
      perror (path);
      return;
    }

  struct stat statbuf;
  if (lstat (path, &statbuf) == -1)
    {
      perror ("fstat");
      return;
    }

  if (S_ISDIR (statbuf.st_mode))
    {
      if (type != 'f')
        {
          printpath (path, name);
        }

      if (depth != 0)
        {
          long newdepth = depth;
          if (depth > 0)
            {
              newdepth--;
            }

          DIR *dir = fdopendir (fd);
          if (dir == NULL)
            {
              die ("fdopendir");
            }

          struct dirent *dirent;
          errno = 0;
          while ((dirent = readdir (dir)) != NULL)
            {
              if (strcmp (dirent->d_name, ".") == 0
                  || strcmp (dirent->d_name, "..") == 0)
                {
                  continue;
                }

              char fullpath[strlen (path) + 1 + strlen (dirent->d_name) + 1];
              strcpy (fullpath, path);
              fullpath[strlen (path)] = '/';
              fullpath[strlen (path) + 1] = '\0';
              strcat (fullpath, dirent->d_name);

              creeper (fullpath, name, type, newdepth);
              errno = 0;
            }

          if (errno != 0)
            {
              die ("readdir");
            }

          if (closedir (dir) == -1)
            {
              die ("closedir");
            }
          return;
        }
    }
  else if (S_ISREG (statbuf.st_mode))
    {
      if (type != 'd')
        {
          printpath (path, name);
        }
    }

  if (close (fd) == -1)
    {
      die ("close");
    }
}

static void
printpath (char *path, char *name)
{
  if (name != NULL)
    {
      char path2[strlen (path) + 1];
      strcpy (path2, path);
      if (fnmatch (name, basename (path), FNM_PERIOD | FNM_PATHNAME) != 0)
        {
          return;
        }
    }

  if (printf ("%s\n", path) < 0)
    {
      fprintf (stderr, "fprintf failed\n");
      exit (EXIT_FAILURE);
    }
}

static void
die (char *msg)
{
  perror (msg);
  exit (EXIT_FAILURE);
}
